#!/usr/bin/perl
no warnings 'deprecated';

# sync two files or two directory trees
$| = 1;

$cflg=0;
$mflg=0;
$oflg=0;
$sflg=0;
$vflg=0;
while(1) {
	if ($ARGV[0] eq "-c") {
		$cflg=1;
		shift;
		next;
	}
	if ($ARGV[0] eq "-m") {
		$mflg=1;
		shift;
		next;
	}
	if ($ARGV[0] eq "-o") {
		$oflg=1;
		shift;
		next;
	}
	if ($ARGV[0] eq "-s") {
		$sflg=1;
		shift;
		next;
	}
	if ($ARGV[0] eq "-v") {
		$vflg=1;
		shift;
		next;
	}
	if ($ARGV[0] =~ /^-/) {
		goto errexit;
	}
	last;
}

$prog=$0;
$prog =~ s,.*/,,;
if ($#ARGV != 1) {
errexit:
	printf STDERR "Usage: %s [-m] [-o] [-s] [-c] [-v] <dir1> <dir2>\n", $prog;
	printf STDERR "\t-m\tIgnore mode/permissions\n";
	printf STDERR "\t-o\tIgnore owner/group\n";
	printf STDERR "\t-s\tSilent about mode setting and errors\n";
	printf STDERR "\t-c\tForce internal copy (do not use UNIX \'cp\')\n";
	printf STDERR "\t-v\tVerbose output\n";
	exit 1;
}

# counters: afound, bfound, updates, deleted, new, mod, own, time
# for files, directories, links, unknown
$afcnt=0; $bfcnt=0; $dfcnt=0; $ufcnt=0; $nfcnt=0; $mfcnt=0; $ofcnt=0; $tfcnt=0;
$adcnt=0; $bdcnt=0; $ddcnt=0; $udcnt=0; $ndcnt=0; $mdcnt=0; $odcnt=0; $tdcnt=0;
$alcnt=0; $blcnt=0; $dlcnt=0; $ulcnt=0; $nlcnt=0;
$aucnt=0; $bucnt=0; $ducnt=0;
# total count
$acnt=0; $bcnt=0;

$winos=($ENV{"OS"} =~ /windows/i);
if ($winos) {
	# Windows forces internal copy and no owner setting
	$cflg=1;
	$oflg=1;
}

$sq="'";
$dq="\"";
@wtype=("none", "symlink", "file", "dir", "other");

if (! -e $ARGV[0]) {
	printf STDERR "%s: %s does not exist\n", $prog, $ARGV[0];
	goto errexit;
}
&syncup($ARGV[0], $ARGV[1]);

# report files
print "Files: found=$afcnt,$bfcnt deleted=$dfcnt new=$nfcnt updated=$ufcnt";
if ($mflg == 0) {
	print " chmod=$mfcnt";
}
if ($oflg == 0) {
	print " chown=$ofcnt";
}
print " utime=$tfcnt\n";
# report directories
print "Directories: found=$adcnt,$bdcnt new=$ndcnt updated=$udcnt deleted=$ddcnt";
if ($mflg == 0) {
	print " chmod=$mdcnt";
}
if ($oflg == 0) {
	print " chown=$odcnt";
}
print " utime=$tdcnt\n";
# report symlinks
unless ($winos) {
	print "Symlinks: found=$alcnt,$blcnt deleted=$dlcnt new=$nlcnt updated=$ulcnt\n";
}
# report other file types
print "Other: found=$aucnt,$bucnt deleted=$ducnt\n";
# report totals
printf "Total: found=%d,%d deleted=%d new=%d updated=%d", $acnt, $bcnt, $dfcnt+$ddcnt+$dlcnt+$ducnt, $nfcnt+$ndcnt+$nlcnt, $ufcnt+$udcnt+$ulcnt;
if ($mflg == 0) {
	printf " chmod=%d", $mfcnt+$mdcnt;
}
if ($oflg == 0) {
	printf " chown=%d", $ofcnt+$odcnt;
}
printf " utime=%d\n", $tfcnt+$tdcnt;
exit 0;

sub syncup() {
	local ($da, $db)=@_;
	local ($la, $lb, $rep, $ma, $mb);
	local (@sa, @sb);
	local (@rb);
	local (%ra);

	# default - change below if discovered different
	$rep="new";

	# STAT 2:mode 4:uid 5:gid 7:size 8:atime 9:mtime 10:ctime
	@sa=lstat($da);
	$ma=&ftype($sa[2]);
	$acnt++;
	if ($ma == 1) {
		$alcnt++;
	} elsif ($ma == 2) {
		$afcnt++;
	} elsif ($ma == 3) {
		$adcnt++;
	} elsif ($ma == 4) {
		$aucnt++;
	} else {
		$acnt--;
	}
	@sb=lstat($db);
	$mb=&ftype($sb[2]);
	$bcnt++;
	if ($mb == 1) {
		$blcnt++;
	} elsif ($mb == 2) {
		$bfcnt++;
	} elsif ($mb == 3) {
		$bdcnt++;
	} elsif ($mb == 4) {
		$bucnt++;
	} else {
		$bcnt--;
	}

	# two files
	if ($ma == 2 && $mb == 2) {
		# compare changed size or mtime (or ctime in UNIX is newer)
		if ($sa[7] == $sb[7] && $sa[9] == $sb[9] &&
			($winos || $sa[10] <= $sb[10])
		) {
chkmo:
			# all is good
			if ($mflg == 0) {
				# check modes
				if (($sa[2] & 07777) != ($sb[2] & 07777)) {
					if ($sflg == 0) {
						printf "%s %s chmod %05o ", $db, $wtype[$mb], ($sa[2] & 07777);
					}
					if (chmod (($sa[2] & 07777), $db)) {
						if ($sflg == 0) {
							printf "OK\n";
						}
						if ($ma == 2) {
							$mfcnt++;
						} else {
							$mdcnt++;
						}
					} else {
						if ($sflg == 0) {
							printf "FAILED: %s", $! . "\n";
						}
					}
				}
			}
			if ($oflg == 0) {
				# check owner/group
				if ($sa[4] != $sb[4] || $sa[5] != $sb[5]) {
					printf "%s %s chown %d,%d ", $db, $wtype[$mb], $sa[4], $sa[5];
					if (chown ($sa[4], $sa[5], $db)) {
						printf "OK\n";
						if ($ma == 2) {
							$ofcnt++;
						} else {
							$odcnt++;
						}
					} else {
						printf "FAILED: %s", $! . "\n";
					}
				}
			}
			return;
		}
		# files are different - delete destination
		unlink $db;
		@sb=();
		$mb=0;
		$rep="updating";
	}

	# two symlinks
	if ($ma == 1 && $mb == 1) {
		$la=readlink($da);
		$lb=readlink($db);
		if ($la eq $lb) {
			return;
		}
		# symlinks are different - delete destination
		unlink $db;
		@sb=();
		$mb=0;
		$rep="updating";
	}

	# two directories
	if ($ma == 3 && $mb == 3) {
twodir:
		# read contents of both directories
		unless (opendir(DA, $da)) {
			printf "%s cannot opendir: %s", $da, $! . "\n";
			return;
		}
		while ($la=readdir(DA)) {
			if ($la eq "." || $la eq "..") {
				next;
			}
			$ra{$la}=1;
		}
		closedir(DA);
		unless (opendir(DB, $db)) {
			printf "%s cannot opendir: %s", $db, $! . "\n";
			return;
		}
		while ($lb=readdir(DB)) {
			if ($lb eq "." || $lb eq "..") {
				next;
			}
			push @rb, $lb;
		}
		closedir(DB);
		foreach $lb (sort @rb) {
			unless ($ra{$lb}) {
				# delete any in B but not in A
				&rmr("$db/$lb");
			}
		}
		foreach $la (sort keys %ra) {
			if ($vflg) {
				print "$db/$la\r";
			}
			# sync all in A
			&syncup("$da/$la", "$db/$la");
			if ($vflg) {
				print $db . " " x (length($la)+1) . "\r";
			}
		}
		# compare mtime of directory
		@sb=lstat($db);
		$mb=&ftype($sb[2]);
		if ($sa[9] != $sb[9]) {
			utime($sb[8], $sa[9], $db);
			$tdcnt++;
			$sb[9]=$sa[9];
		}
		goto chkmo;
	}

	# mismatched type
	if ($ma != $mb && $#sb > 1) {
		printf "%s is a %s, but %s is a %s\n", $da, $wtype[$ma], $db, $wtype[$mb];
		if (&rmr($db) == 0) {
			return;
		}
		# WARNING: it might not be deleted
		@sb=lstat($db);
		$mb=&ftype($sb[2]);
	}

	# new thing (or update)
	if ($mb == 0) {
		if ($ma == 1) {
			$la=readlink($da);
			printf "%s symlink %s to %s", $db, $rep, $la;
			if (symlink($la, $db)) {
				printf "OK\n";
			} else {
				printf "FAILED: %s", $! . "\n";
			}
			if ($rep eq "new") {
				$nlcnt++;
			} else {
				$ulcnt++;
			}
			return;
		}
		if ($ma == 2) {
			printf "%s %s ", $db, $rep;
			if ($cflg == 1) {
				unless (open(FA, "<$da")) {
					printf "%s cannot open/read: %s", $da, $! . "\n";
					return;
				}
				unless (open(FB, ">$db")) {
					printf "%s cannot open/write: %s", $db, $! . "\n";
					close FA;
					return;
				}
				binmode FA;
				binmode FB;
				$wc=0;
				$lop="";
				while ($l=<FA>) {
					print FB $l;
					$wc += length($l);
					if ($sa[7] > 10*1024*1024) {
						$op=sprintf "%dM/%dM", int($wc/1024/1024 + 0.5), int($sa[7]/1024/1024 + 0.5);
					} else {
						$op=sprintf "%dk/%dk", int($wc/1024 + 0.5), int($sa[7]/1024 + 0.5);
					}
					if ($op ne $lop) {
						print "$op" . "\b" x length($op);
						$lop=$op;
					}
				}
				close FA;
				close FB;
				if ($lop) {
					print " " x length($lop) . "\b" x length($lop);
				}
				print $wc;
			} else {
				$qda=$da;
				$qda =~ s/$sq/$sq$dq$sq$dq$sq/g;
				$qdb=$db;
				$qdb =~ s/$sq/$sq$dq$sq$dq$sq/g;
				$wc=$sa[7];
				print $wc;
				system("cp -p '$qda' '$qdb'");
			}
			# NOW get info of files - it may have changed
			@sa=lstat($da);
			$ma=&ftype($sa[2]);
			@sb=lstat($db);
			$mb=&ftype($sb[2]);
			# did copy size match
			if ($sa[7] != $sb[7]) {
				print " (", $sa[7], " != ", $sb[7], ")";
			}
			printf "\n";
			if ($rep eq "new") {
				$nfcnt++;
			} else {
				$ufcnt++;
			}
			# compare mtime of file
			if ($sa[9] != $sb[9]) {
				utime($sb[8], $sa[9], $db);
				$tfcnt++;
				$sb[9]=$sa[9];
			}
			goto chkmo;
		}
		if ($ma == 3) {
			printf "%s mkdir\n", $db;
			unless (mkdir($db)) {
				printf "$db mkdir fail: %s", $! . "\n";
				return;
			}
			if ($rep eq "new") {
				$ndcnt++;
			} else {
				$udcnt++;
			}
			@sb=lstat($db);
			$mb=&ftype($sb[2]);
			goto twodir;
		}
	}

	if ($ma == 4) {
		# type is unknown
		printf "%s unknown file type - ignored\n", $da;
	}
}

# file type (ordered by statistical expectation of discovery)
sub ftype() {
	local ($m)=@_;

	if (($m & 0170000) == 0100000) {
		# a file
		return 2;
	}
	if (($m & 0170000) == 0040000) {
		# a directory
		return 3;
	}
	if ($m == 0) {
		# no such file
		return 0;
	}
	if (($m & 0170000) == 0120000) {
		# a symlink
		return 1;
	}
	# other
	return 4;
}

# recursive remove
sub rmr() {
	local ($p)=@_;
	local ($f, $m);
	local (@a);

	# directory
	@a=lstat($p);
	$m=&ftype($a[2]);
	if ($m == 3) {
		unless (opendir(D, $p)) {
			printf "%s cannot opendir: %s", $p, $! . "\n";
			return 0;
		}
		@a=();
		while ($f=readdir(D)) {
			if ($f eq "." || $f eq "..") {
				next;
			}
			push @a, $f;
		}
		closedir D;
		foreach $f (sort @a) {
			unless (&rmr("$p/$f")) {
				return 0;
			}
		}
		unless (rmdir($p)) {
			printf "%s not deleted: %s", $p, $! . "\n";
			return 0;
		}
		$ddcnt++;
		printf "%s rmdir\n", $p;
		return 1;
	}

	# not there?
	if ($m == 0) {
		printf "%s not found\n", $p;
		return 0;
	}

	# non directories
	$f=$wtype[$m];
	unless (unlink($p)) {
		printf "%s %s not deleted\n", $p, $f;
		return 0;
	}
	printf "%s %s deleted\n", $p, $f;
	if ($m == 2) {
		$dfcnt++;
	} elsif ($m == 1) {
		$dlcnt++;
	} else {
		$ducnt++;
	}
	return 1;
}
